@@ -254,8 +254,8 @@ def mark_as_titlepage(container, name, move_to_start=True):
 def find_cover_page(container):
     'Find a document marked as a cover in the OPF'
     ver = container.opf_version_parsed
+    mm = container.mime_map
     if ver.major < 3:
-        mm = container.mime_map
         guide_type_map = container.guide_type_map
         for ref_type, name in guide_type_map.iteritems():
             if ref_type.lower() == 'cover' and mm.get(name, '').lower() in OEB_DOCS:
@@ -263,6 +263,24 @@ def find_cover_page(container):
     else:
         for name in container.manifest_items_with_property('calibre:title-page'):
             return name
+        from calibre.ebooks.oeb.polish.toc import get_landmarks
+        for landmark in get_landmarks(container):
+            if landmark['type'] == 'cover' and mm.get(landmark['dest'], '').lower() in OEB_DOCS:
+                return landmark['dest']
+
+
+def fix_conversion_titlepage_links_in_nav(container):
+    from calibre.ebooks.oeb.polish.toc import find_existing_nav_toc
+    cover_page_name = find_cover_page(container)
+    if not cover_page_name:
+        return
+    nav_page_name = find_existing_nav_toc(container)
+    if not nav_page_name:
+        return
+    for elem in container.parsed(nav_page_name).xpath('//*[@data-calibre-removed-titlepage]'):
+        elem.attrib.pop('data-calibre-removed-titlepage')
+        elem.set('href', container.name_to_href(cover_page_name, nav_page_name))
+    container.dirty(nav_page_name)
 
 
 def find_cover_image_in_page(container, cover_page):
@@ -302,9 +320,11 @@ def clean_opf(container):
             name = gtm.get(typ, None)
             if name and name in container.name_path_map:
                 yield name
-    removed_names = container.apply_unique_properties(None, 'cover-image', 'calibre:title-page')[0]
-    for name in removed_names:
-        yield name
+    ver = container.opf_version_parsed
+    if ver.major > 2:
+        removed_names = container.apply_unique_properties(None, 'cover-image', 'calibre:title-page')[0]
+        for name in removed_names:
+            yield name
     container.dirty(container.opf_name)
 
 
