@@ -19,6 +19,9 @@ from calibre.gui2 import config, error_dialog, question_dialog, gprefs
 from calibre.gui2.dialogs.confirm_delete import confirm
 from calibre.gui2.dialogs.saved_search_editor import SavedSearchEditor
 from calibre.gui2.dialogs.search import SearchDialog
+from calibre.utils.icu import primary_sort_key
+
+QT_HIDDEN_CLEAR_ACTION = '_q_qlineeditclearaction'
 
 
 class AsYouType(unicode):
@@ -102,11 +105,16 @@ class SearchBox2(QComboBox):  # {{{
     changed = pyqtSignal()
     focus_to_library = pyqtSignal()
 
-    def __init__(self, parent=None):
+    def __init__(self, parent=None, add_clear_action=True):
         QComboBox.__init__(self, parent)
         self.normal_background = 'rgb(255, 255, 255, 0%)'
         self.line_edit = SearchLineEdit(self)
         self.setLineEdit(self.line_edit)
+        if add_clear_action:
+            self.lineEdit().setClearButtonEnabled(True)
+            ac = self.findChild(QAction, QT_HIDDEN_CLEAR_ACTION)
+            if ac is not None:
+                ac.triggered.connect(self.clear_clicked)
 
         c = self.line_edit.completer()
         c.setCompletionMode(c.PopupCompletion)
@@ -127,6 +135,11 @@ class SearchBox2(QComboBox):  # {{{
         self._in_a_search = False
         self.tool_tip_text = self.toolTip()
 
+    def add_action(self, icon, position=QLineEdit.TrailingPosition):
+        if not isinstance(icon, QIcon):
+            icon = QIcon(I(icon))
+        return self.lineEdit().addAction(icon, position)
+
     def initialize(self, opt_name, colorize=False, help_text=_('Search')):
         self.as_you_type = config['search_as_you_type']
         self.opt_name = opt_name
@@ -166,6 +179,7 @@ class SearchBox2(QComboBox):  # {{{
 
     def clear_clicked(self, *args):
         self.clear()
+        self.setFocus(Qt.OtherFocusReason)
 
     def search_done(self, ok):
         if isinstance(ok, basestring):
@@ -203,12 +217,15 @@ class SearchBox2(QComboBox):  # {{{
         if k in (Qt.Key_Enter, Qt.Key_Return):
             return self.do_search()
         if k not in (Qt.Key_Up, Qt.Key_Down):
-            QComboBox.keyPressEvent(self, event)
+            return QComboBox.keyPressEvent(self, event)
+        self.blockSignals(True)
+        self.normalize_state()
+        if k == Qt.Key_Down and self.currentIndex() == 0 and not self.lineEdit().text():
+            self.setCurrentIndex(1), self.setCurrentIndex(0)
+            event.accept()
         else:
-            self.blockSignals(True)
-            self.normalize_state()
             QComboBox.keyPressEvent(self, event)
-            self.blockSignals(False)
+        self.blockSignals(False)
 
     def completer_used(self, text):
         self.timer.stop()
@@ -437,14 +454,13 @@ class SearchBoxMixin(object):  # {{{
 
     def init_search_box_mixin(self):
         self.search.initialize('main_search_history', colorize=True,
-                help_text=_('Search (For Advanced search click the button to the left)'))
+                help_text=_('Search (For advanced search click the gear icon to the left)'))
         self.search.cleared.connect(self.search_box_cleared)
         # Queued so that search.current_text will be correct
         self.search.changed.connect(self.search_box_changed,
                 type=Qt.QueuedConnection)
         self.search.focus_to_library.connect(self.focus_to_library)
-        self.clear_button.clicked.connect(self.search.clear_clicked)
-        self.advanced_search_button.clicked[bool].connect(self.do_advanced_search)
+        self.advanced_search_toggle_action.triggered.connect(self.do_advanced_search)
 
         self.search.clear()
         self.search.setMaximumWidth(self.width()-150)
@@ -459,8 +475,6 @@ class SearchBoxMixin(object):  # {{{
         self.addAction(self.action_focus_search)
         self.search.setStatusTip(re.sub(r'<\w+>', ' ',
             unicode(self.search.toolTip())))
-        self.advanced_search_button.setStatusTip(self.advanced_search_button.toolTip())
-        self.clear_button.setStatusTip(self.clear_button.toolTip())
         self.set_highlight_only_button_icon()
         self.highlight_only_button.clicked.connect(self.highlight_only_clicked)
         tt = _('Enable or disable search highlighting.') + '<br><br>'
@@ -482,10 +496,13 @@ class SearchBoxMixin(object):  # {{{
         self.focus_to_library()
 
     def set_highlight_only_button_icon(self):
+        b = self.highlight_only_button
         if config['highlight_search_matches']:
-            self.highlight_only_button.setIcon(QIcon(I('highlight_only_on.png')))
+            b.setIcon(QIcon(I('highlight_only_on.png')))
+            b.setText(_('Filter'))
         else:
-            self.highlight_only_button.setIcon(QIcon(I('highlight_only_off.png')))
+            b.setIcon(QIcon(I('highlight_only_off.png')))
+            b.setText(_('Highlight'))
         self.highlight_only_button.setVisible(gprefs['show_highlight_toggle_button'])
         self.library_view.model().set_highlight_only(config['highlight_search_matches'])
 
@@ -524,14 +541,16 @@ class SavedSearchBoxMixin(object):  # {{{
 
     def init_saved_seach_box_mixin(self):
         self.saved_search.changed.connect(self.saved_searches_changed)
-        self.clear_button.clicked.connect(self.saved_search.clear)
+        ac = self.search.findChild(QAction, QT_HIDDEN_CLEAR_ACTION)
+        if ac is not None:
+            ac.triggered.connect(self.saved_search.clear)
         self.save_search_button.clicked.connect(
                                 self.saved_search.save_search_button_clicked)
         self.copy_search_button.clicked.connect(
                                 self.saved_search.copy_search_button_clicked)
         # self.saved_searches_changed()
         self.saved_search.initialize(self.search, colorize=True,
-                help_text=_('Saved Searches'))
+                help_text=_('Saved searches'))
         self.saved_search.tool_tip_text=_('Choose saved search or enter name for new saved search')
         self.saved_search.setToolTip(self.saved_search.tool_tip_text)
         self.saved_search.setStatusTip(self.saved_search.tool_tip_text)
@@ -544,12 +563,26 @@ class SavedSearchBoxMixin(object):  # {{{
         self.save_search_button.setMenu(QMenu())
         self.save_search_button.menu().addAction(
                             QIcon(I('plus.png')),
-                            _('Create saved search'),
+                            _('Create Saved search'),
                             self.saved_search.save_search_button_clicked)
         self.save_search_button.menu().addAction(
-            QIcon(I('trash.png')), _('Delete saved search'), self.saved_search.delete_current_search)
+            QIcon(I('trash.png')), _('Delete Saved search'), self.saved_search.delete_current_search)
         self.save_search_button.menu().addAction(
-            QIcon(I('search.png')), _('Manage saved searches'), partial(self.do_saved_search_edit, None))
+            QIcon(I('search.png')), _('Manage Saved searches'), partial(self.do_saved_search_edit, None))
+        self.add_saved_search_button.setMenu(QMenu())
+        self.add_saved_search_button.menu().aboutToShow.connect(self.populate_add_saved_search_menu)
+
+    def populate_add_saved_search_menu(self):
+        m = self.add_saved_search_button.menu()
+        m.clear()
+        m.addAction(QIcon(I('plus.png')), _('Add Saved search'), self.add_saved_search)
+        m.addAction(QIcon(I("search_copy_saved.png")), _('Get Saved search expression'),
+                    self.get_saved_search_text)
+        m.addActions(list(self.save_search_button.menu().actions())[-1:])
+        m.addSeparator()
+        db = self.current_db
+        for name in sorted(db.saved_search_names(), key=lambda x: primary_sort_key(x.strip())):
+            m.addAction(name.strip(), partial(self.saved_search.saved_search_selected, name))
 
     def saved_searches_changed(self, set_restriction=None, recount=True):
         self.build_search_restriction_list()
@@ -568,4 +601,27 @@ class SavedSearchBoxMixin(object):  # {{{
         self.saved_searches_changed()
         self.saved_search.clear()
 
+    def add_saved_search(self):
+        from calibre.gui2.dialogs.saved_search_editor import AddSavedSearch
+        d = AddSavedSearch(parent=self, search=self.search.current_text)
+        if d.exec_() == d.Accepted:
+            self.current_db.new_api.ensure_has_search_category(fail_on_existing=False)
+            self.do_rebuild_saved_searches()
+
+    def get_saved_search_text(self):
+        db = self.current_db
+        try:
+            current_search = self.search.currentText()
+            if not current_search.startswith('search:'):
+                raise ValueError()
+            # This strange expression accounts for the four ways a search can be written:
+            # search:fff, search:"fff", search:"=fff". and search:="fff"
+            current_search = current_search[7:].lstrip('=').strip('"').lstrip('=')
+            current_search = db.saved_search_lookup(current_search)
+            if not current_search:
+                raise ValueError()
+            self.search.set_search_string(current_search)
+        except:
+            from calibre.gui2.ui import get_gui
+            get_gui().status_bar.show_message(_('Current search is not a saved search'), 3000)
     # }}}
